/*
 * Copyright (c) 2009, The MilkyTracker Team.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 * - Neither the name of the <ORGANIZATION> nor the names of its contributors
 *   may be used to endorse or promote products derived from this software
 *   without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*
 *  ResamplerCubic.h
 *  MilkyPlay
 *
 *  Created by Peter Barth on 03.01.08.
 *
 */

/*
 * Cubic 4 Point 3rd order polynomial interpolation resampler                 
 *
 * Formulas from:
 * Polynomial Interpolators for High-Quality Resampling of Oversampled Audio by 
 * Olli Niemitalo ollinie@freenet.hut.fi 
 *
 * Website: http://www.student.oulu.fi/~oniemita/DSP/INDEX.HTM
 *
 */
 
#define __DEIP__
#define fpmul MP_FP_MUL

enum CubicResamplers
{
	CubicResamplerLagrange,
	CubicResamplerSpline
};

// here is some code to work around my VS6.1 backward compability
template<bool ramping, CubicResamplers type, class bufferType, mp_uint32 shift>
class CubicResamplerDummy
{
public:
	static inline mp_sint32 interpolate_lagrange4Point(const bufferType* sample,
													   const mp_sint32 smppos)
	{
		// note that the sample buffer is padded with leading and
		// trailing space which is filled with samples from the 
		// beginning/end, so even if it SEEMS we are crossing sample
		// boundaries by -1, +1 and +2, it doesn't happen
#ifdef __DEIP__		
		const mp_sint32 v1 = (mp_sint32)sample[(smppos>>16)] << (16-shift);
		const mp_sint32 v2 = (mp_sint32)sample[(smppos>>16) + 1] << (16-shift);							
		const mp_sint32 v0 = (mp_sint32)sample[(smppos>>16) - 1] << (16-shift);
		const mp_sint32 v3 = (mp_sint32)sample[(smppos>>16) + 2] << (16-shift);
		
		const mp_sint32 x = smppos & 65535;
		
		mp_sint32 c0 = v1; 
		mp_sint32 c1 = v2 - ((v0*(65536/3))>>16) - ((v3*(65536/6))>>16) - (v1>>1); 
		mp_sint32 c2 = ((v0+v2)>>1) - v1; 
		mp_sint32 c3 = (((65536/6)*(v3-v0))>>16) + ((v1-v2)>>1); 
		
		return fpmul(fpmul(fpmul(c3,x)+c2,x)+c1, x)+c0;
#else
		const mp_sint32 v1 = sample[(smppos>>16)] << (16-shift);
		const mp_sint32 v2 = sample[(smppos>>16) + 1] << (16-shift);							
		const mp_sint32 v0 = sample[(smppos>>16) - 1] << (16-shift);
		mp_sint32 v3 = sample[(smppos>>16) + 2] << 8;
		const mp_sint32 ofsf = (smppos&65535) + 65536;
		v3 += -3*v2 + 3*v1 - v0;
		v3 = fpmul(v3, ((ofsf - 2*65536) * (65536/6)) >> 16);
		v3 += v2 - v1 - v1 + v0;
		v3 = fpmul(v3, (ofsf - 65536) >> 1);
		v3 += v1 - v0;
		v3 = fpmul(v3, ofsf);
		v3 += v0;	
		return v3;
#endif
	}
	
	static inline mp_sint32 interpolate_spline4Point(const bufferType* sample,
													 const mp_sint32 smppos)
	{
#ifdef __DEIP__
		const mp_sint32 v1 = sample[(smppos>>16)] << (16-shift);
		const mp_sint32 v2 = sample[(smppos>>16) + 1] << (16-shift);							
		const mp_sint32 v0 = sample[(smppos>>16) - 1] << (16-shift);
		const mp_sint32 v3 = sample[(smppos>>16) + 2] << (16-shift);
		
		const mp_sint32 x = smppos & 65535;
		
		mp_sint32 ym1py1 = v0+v2; 
		mp_sint32 c0 = (((65536/6)*ym1py1) + (65536*2/3)*v1) >> 16; 
		mp_sint32 c1 = (v2-v0)>>1; 
		mp_sint32 c2 = (ym1py1>>1) - v1; 
		mp_sint32 c3 = ((v1-v2)>>1) + (((65536/6)*(v3-v0)) >> 16); 
		
		return fpmul(fpmul(fpmul(c3,x)+c2,x)+c1, x)+c0;
#else
		// this the stuff timidity uses, I have no clue what it does, but
		// it causes heavy gibbs phenomena and doesn't look nearly as the above
		const mp_sint32 v1 = sample[(smppos>>16)] << 8;
		const mp_sint32 v2 = sample[(smppos>>16) + 1] << 8;							
		const mp_sint32 v0 = sample[(smppos>>16) - 1] << 8;
		const mp_sint32 v3 = sample[(smppos>>16) + 2] << 8;
		
		const mp_sint32 ofsf = smppos & 65535;
		const mp_sint32 temp = v2;
		const mp_sint32 temp2 = (5 * v3 - 11 * v2 + 7 * v1 - v0) >> 2;
		
		const mp_sint32 hres = fpmul(fpmul(fpmul(temp2, ofsf + 65536), (ofsf - 65536)) + 6*v2, ofsf);
		
		const mp_sint32 temp3 = (5 * v0 - 11 * v1 + 7 * temp - v3) >> 2;
		
		const mp_sint32 temp4 = fpmul(temp3, ofsf) >> 16;
		const mp_sint32 temp5 = fpmul(temp4, ofsf - 65536*2) >> 16;
		const mp_sint32 res = fpmul(6*v1+temp5, 65536 - ofsf);
		
		return fpmul(res + hres, 65536 / 6);
#endif
	}
	
	static inline void addBlock(mp_sint32* buffer, ChannelMixer::TMixerChannel* chn, mp_uint32 count)
	{
		// adding some local variables, will be faster to access than attributes of chn
		mp_sint32 voll = chn->finalvoll;
		mp_sint32 volr = chn->finalvolr;
		
		const mp_sint32 rampFromVolStepL = ramping ? chn->rampFromVolStepL : 0;
		const mp_sint32 rampFromVolStepR = ramping ? chn->rampFromVolStepR : 0;		
		
		mp_sint32 smppos = chn->smppos;
		mp_sint32 smpposfrac = chn->smpposfrac;
		const mp_sint32 smpadd = (chn->flags&ChannelMixer::MP_SAMPLE_BACKWARD) ? -chn->smpadd : chn->smpadd;
		
		mp_sint32 fp = smpadd*count;
		MP_INCREASESMPPOS(chn->smppos,chn->smpposfrac,fp,16);
		
		const bufferType* sample = ((const bufferType*)chn->sample) + smppos;
		smppos = smpposfrac;
		
		while (count--)
		{
			mp_sint32 s;
			
			switch (type)
			{
				case CubicResamplerLagrange:
					s = interpolate_lagrange4Point(sample, smppos);
					break;
				case CubicResamplerSpline:
					s = interpolate_spline4Point(sample, smppos);
					break;
			}
			
			(*buffer++)+=(s*(voll>>15))>>15; 
			(*buffer++)+=(s*(volr>>15))>>15; 
			
			// ramping is a template parameter
			// see explanation above
			if (ramping)
			{
				voll+=rampFromVolStepL; 
				volr+=rampFromVolStepR; 
			}
			
			smppos+=smpadd;
		}
		
		if (ramping)
		{
			chn->finalvoll = voll;
			chn->finalvolr = volr;	
		}
	}
};

template<bool ramping, CubicResamplers type>
class ResamplerLagrange : public ChannelMixer::ResamplerBase
{
public:
	virtual bool isRamping() { return ramping; }
	virtual bool supportsFullChecking() { return false; }
	virtual bool supportsNoChecking() { return true; }

	virtual void addBlockNoCheck(mp_sint32* buffer, ChannelMixer::TMixerChannel* chn, mp_uint32 count)
	{
		if (chn->flags & 4)
			CubicResamplerDummy<ramping, type, mp_sword, 16>::addBlock(buffer, chn, count);		
		else
			CubicResamplerDummy<ramping, type, mp_sbyte, 8>::addBlock(buffer, chn, count);
	}
};


static const int gauss[]={
	0x000, 0x000, 0x000, 0x000, 0x000, 0x000, 0x000, 0x000, 
	0x000, 0x000, 0x000, 0x000, 0x000, 0x000, 0x000, 0x000, 
	0x001, 0x001, 0x001, 0x001, 0x001, 0x001, 0x001, 0x001, 
	0x001, 0x001, 0x001, 0x002, 0x002, 0x002, 0x002, 0x002, 
	0x002, 0x002, 0x003, 0x003, 0x003, 0x003, 0x003, 0x004, 
	0x004, 0x004, 0x004, 0x004, 0x005, 0x005, 0x005, 0x005, 
	0x006, 0x006, 0x006, 0x006, 0x007, 0x007, 0x007, 0x008, 
	0x008, 0x008, 0x009, 0x009, 0x009, 0x00A, 0x00A, 0x00A, 
	0x00B, 0x00B, 0x00B, 0x00C, 0x00C, 0x00D, 0x00D, 0x00E, 
	0x00E, 0x00F, 0x00F, 0x00F, 0x010, 0x010, 0x011, 0x011, 
	0x012, 0x013, 0x013, 0x014, 0x014, 0x015, 0x015, 0x016, 
	0x017, 0x017, 0x018, 0x018, 0x019, 0x01A, 0x01B, 0x01B, 
	0x01C, 0x01D, 0x01D, 0x01E, 0x01F, 0x020, 0x020, 0x021, 
	0x022, 0x023, 0x024, 0x024, 0x025, 0x026, 0x027, 0x028, 
	0x029, 0x02A, 0x02B, 0x02C, 0x02D, 0x02E, 0x02F, 0x030, 
	0x031, 0x032, 0x033, 0x034, 0x035, 0x036, 0x037, 0x038, 
	0x03A, 0x03B, 0x03C, 0x03D, 0x03E, 0x040, 0x041, 0x042, 
	0x043, 0x045, 0x046, 0x047, 0x049, 0x04A, 0x04C, 0x04D, 
	0x04E, 0x050, 0x051, 0x053, 0x054, 0x056, 0x057, 0x059, 
	0x05A, 0x05C, 0x05E, 0x05F, 0x061, 0x063, 0x064, 0x066, 
	0x068, 0x06A, 0x06B, 0x06D, 0x06F, 0x071, 0x073, 0x075, 
	0x076, 0x078, 0x07A, 0x07C, 0x07E, 0x080, 0x082, 0x084, 
	0x086, 0x089, 0x08B, 0x08D, 0x08F, 0x091, 0x093, 0x096, 
	0x098, 0x09A, 0x09C, 0x09F, 0x0A1, 0x0A3, 0x0A6, 0x0A8, 
	0x0AB, 0x0AD, 0x0AF, 0x0B2, 0x0B4, 0x0B7, 0x0BA, 0x0BC, 
	0x0BF, 0x0C1, 0x0C4, 0x0C7, 0x0C9, 0x0CC, 0x0CF, 0x0D2, 
	0x0D4, 0x0D7, 0x0DA, 0x0DD, 0x0E0, 0x0E3, 0x0E6, 0x0E9, 
	0x0EC, 0x0EF, 0x0F2, 0x0F5, 0x0F8, 0x0FB, 0x0FE, 0x101, 
	0x104, 0x107, 0x10B, 0x10E, 0x111, 0x114, 0x118, 0x11B, 
	0x11E, 0x122, 0x125, 0x129, 0x12C, 0x130, 0x133, 0x137, 
	0x13A, 0x13E, 0x141, 0x145, 0x148, 0x14C, 0x150, 0x153, 
	0x157, 0x15B, 0x15F, 0x162, 0x166, 0x16A, 0x16E, 0x172, 
	0x176, 0x17A, 0x17D, 0x181, 0x185, 0x189, 0x18D, 0x191, 
	0x195, 0x19A, 0x19E, 0x1A2, 0x1A6, 0x1AA, 0x1AE, 0x1B2, 
	0x1B7, 0x1BB, 0x1BF, 0x1C3, 0x1C8, 0x1CC, 0x1D0, 0x1D5, 
	0x1D9, 0x1DD, 0x1E2, 0x1E6, 0x1EB, 0x1EF, 0x1F3, 0x1F8, 
	0x1FC, 0x201, 0x205, 0x20A, 0x20F, 0x213, 0x218, 0x21C, 
	0x221, 0x226, 0x22A, 0x22F, 0x233, 0x238, 0x23D, 0x241, 
	0x246, 0x24B, 0x250, 0x254, 0x259, 0x25E, 0x263, 0x267, 
	0x26C, 0x271, 0x276, 0x27B, 0x280, 0x284, 0x289, 0x28E, 
	0x293, 0x298, 0x29D, 0x2A2, 0x2A6, 0x2AB, 0x2B0, 0x2B5, 
	0x2BA, 0x2BF, 0x2C4, 0x2C9, 0x2CE, 0x2D3, 0x2D8, 0x2DC, 
	0x2E1, 0x2E6, 0x2EB, 0x2F0, 0x2F5, 0x2FA, 0x2FF, 0x304, 
	0x309, 0x30E, 0x313, 0x318, 0x31D, 0x322, 0x326, 0x32B, 
	0x330, 0x335, 0x33A, 0x33F, 0x344, 0x349, 0x34E, 0x353, 
	0x357, 0x35C, 0x361, 0x366, 0x36B, 0x370, 0x374, 0x379, 
	0x37E, 0x383, 0x388, 0x38C, 0x391, 0x396, 0x39B, 0x39F, 
	0x3A4, 0x3A9, 0x3AD, 0x3B2, 0x3B7, 0x3BB, 0x3C0, 0x3C5, 
	0x3C9, 0x3CE, 0x3D2, 0x3D7, 0x3DC, 0x3E0, 0x3E5, 0x3E9, 
	0x3ED, 0x3F2, 0x3F6, 0x3FB, 0x3FF, 0x403, 0x408, 0x40C, 
	0x410, 0x415, 0x419, 0x41D, 0x421, 0x425, 0x42A, 0x42E, 
	0x432, 0x436, 0x43A, 0x43E, 0x442, 0x446, 0x44A, 0x44E, 
	0x452, 0x455, 0x459, 0x45D, 0x461, 0x465, 0x468, 0x46C, 
	0x470, 0x473, 0x477, 0x47A, 0x47E, 0x481, 0x485, 0x488, 
	0x48C, 0x48F, 0x492, 0x496, 0x499, 0x49C, 0x49F, 0x4A2, 
	0x4A6, 0x4A9, 0x4AC, 0x4AF, 0x4B2, 0x4B5, 0x4B7, 0x4BA, 
	0x4BD, 0x4C0, 0x4C3, 0x4C5, 0x4C8, 0x4CB, 0x4CD, 0x4D0, 
	0x4D2, 0x4D5, 0x4D7, 0x4D9, 0x4DC, 0x4DE, 0x4E0, 0x4E3, 
	0x4E5, 0x4E7, 0x4E9, 0x4EB, 0x4ED, 0x4EF, 0x4F1, 0x4F3, 
	0x4F5, 0x4F6, 0x4F8, 0x4FA, 0x4FB, 0x4FD, 0x4FF, 0x500, 
	0x502, 0x503, 0x504, 0x506, 0x507, 0x508, 0x50A, 0x50B, 
	0x50C, 0x50D, 0x50E, 0x50F, 0x510, 0x511, 0x511, 0x512, 
	0x513, 0x514, 0x514, 0x515, 0x516, 0x516, 0x517, 0x517, 
	0x517, 0x518, 0x518, 0x518, 0x518, 0x518, 0x519, 0x519};

static const int	*G1 = &gauss[256];
static const int	*G2 = &gauss[512];
static const int	*G3 = &gauss[255];
static const int	*G4 = &gauss[-1];

template<bool ramping, class bufferType, mp_uint32 shift>
class GaussResamplerDummy
{
public:
	static inline mp_sint32 interpolate_gauss4Point(const bufferType* sample,
													   const mp_sint32 smppos)
	{
		// note that the sample buffer is padded with leading and
		// trailing space which is filled with samples from the 
		// beginning/end, so even if it SEEMS we are crossing sample
		// boundaries by -1, +1 and +2, it doesn't happen
#ifdef __DEIP__		
		const mp_sint32 v1 = (mp_sint32)sample[(smppos>>16)] << (16-shift);
		const mp_sint32 v2 = (mp_sint32)sample[(smppos>>16) + 1] << (16-shift);							
		const mp_sint32 v0 = (mp_sint32)sample[(smppos>>16) - 1] << (16-shift);
		const mp_sint32 v3 = (mp_sint32)sample[(smppos>>16) + 2] << (16-shift);
		
		const mp_sint32 x = ((smppos & 65535) >> 8)-256;

		int	out = 0;
		out = ( ( G4[ -x ] * v0 ) >> 11 ) & ~1;
		out += ( ( G3[ -x ] * v1 ) >> 11 ) & ~1;
		out += ( ( G2[ x ] * v2 ) >> 11 ) & ~1;
		out += ( ( G1[ x ] * v3 ) >> 11 ) & ~1;
		
		if ( out > 32767 ) {
			out = 32767;
		}
		else if ( out < -32768 ) {
			out = -32768;
		}
		return out;
#else
		const mp_sint32 v1 = sample[(smppos>>16)] << (16-shift);
		const mp_sint32 v2 = sample[(smppos>>16) + 1] << (16-shift);							
		const mp_sint32 v0 = sample[(smppos>>16) - 1] << (16-shift);
		mp_sint32 v3 = sample[(smppos>>16) + 2] << 8;
		const mp_sint32 ofsf = (smppos&65535) + 65536;
		v3 += -3*v2 + 3*v1 - v0;
		v3 = fpmul(v3, ((ofsf - 2*65536) * (65536/6)) >> 16);
		v3 += v2 - v1 - v1 + v0;
		v3 = fpmul(v3, (ofsf - 65536) >> 1);
		v3 += v1 - v0;
		v3 = fpmul(v3, ofsf);
		v3 += v0;	
		return v3;
#endif
	}
	
	static inline void addBlock(mp_sint32* buffer, ChannelMixer::TMixerChannel* chn, mp_uint32 count)
	{
		// adding some local variables, will be faster to access than attributes of chn
		mp_sint32 voll = chn->finalvoll;
		mp_sint32 volr = chn->finalvolr;
		
		const mp_sint32 rampFromVolStepL = ramping ? chn->rampFromVolStepL : 0;
		const mp_sint32 rampFromVolStepR = ramping ? chn->rampFromVolStepR : 0;		
		
		mp_sint32 smppos = chn->smppos;
		mp_sint32 smpposfrac = chn->smpposfrac;
		const mp_sint32 smpadd = (chn->flags&ChannelMixer::MP_SAMPLE_BACKWARD) ? -chn->smpadd : chn->smpadd;
		
		mp_sint32 fp = smpadd*count;
		MP_INCREASESMPPOS(chn->smppos,chn->smpposfrac,fp,16);
		
		const bufferType* sample = ((const bufferType*)chn->sample) + smppos;
		smppos = smpposfrac;
		
		while (count--)
		{
			mp_sint32 s;
			
			s = interpolate_gauss4Point(sample, smppos);
			
			(*buffer++)+=(s*(voll>>15))>>15; 
			(*buffer++)+=(s*(volr>>15))>>15; 
			
			// ramping is a template parameter
			// see explanation above
			if (ramping)
			{
				voll+=rampFromVolStepL; 
				volr+=rampFromVolStepR; 
			}
			
			smppos+=smpadd;
		}
		
		if (ramping)
		{
			chn->finalvoll = voll;
			chn->finalvolr = volr;	
		}
	}
};

template<bool ramping>
class ResamplerGauss : public ChannelMixer::ResamplerBase
{
public:
	virtual bool isRamping() { return ramping; }
	virtual bool supportsFullChecking() { return false; }
	virtual bool supportsNoChecking() { return true; }
	
	virtual void addBlockNoCheck(mp_sint32* buffer, ChannelMixer::TMixerChannel* chn, mp_uint32 count)
	{
		if (chn->flags & 4)
			GaussResamplerDummy<ramping, mp_sword, 16>::addBlock(buffer, chn, count);		
		else
			GaussResamplerDummy<ramping, mp_sbyte, 8>::addBlock(buffer, chn, count);
	}
};

#undef __DEIP__
#undef fpmul
